@@ -9,7 +9,7 @@
  * The copyright holder will however retain the copyright.
  * In addition to vpnd this code is put in the 'public domain' (*)
  * for all GPL/LGPL based open source software products.
- * For all other software products please contact astmail@yahoo.com
+ * For all other software products please contact ast@domdv.de
  * for a license agreement. There is no guarantee for the fitness
  * and usability of this code for any purpose. The author takes no
  * responsibility for any damages caused by the use of this code.
@@ -46,10 +46,22 @@
 #undef WORD32
 #endif
 
+#ifdef WORD64
+#undef WORD64
+#endif
+
+#ifdef BITS16
+#undef BITS16
+#endif
+
 #ifdef BITS32
 #undef BITS32
 #endif
 
+#ifdef BITS64
+#undef BITS64
+#endif
+
 /* make sure unsigned char is 8 bits wide */
 
 #if UCHAR_MAX == 0xff
@@ -58,16 +70,28 @@
 #error No 8 bit value available
 #endif
 
-/* make sure that WORD16 is at least 16 bits wide */
+/* make sure that WORD16 is at least 16 bits wide,
+   default the macro BITS16 to nothing if WORD16
+   is exactly 16 bits wide */
 
 #if USHRT_MAX == 0xffff
 #define WORD16 unsigned short
+#define BITS16(a) (a)
+#elif UINT_MAX == 0xffff
+#define WORD16 unsigned int
+#define BITS16(a) (a)
 #elif ULONG_MAX == 0xffff
 #define WORD16 unsigned long
+#define BITS16(a) (a)
 #elif USHRT_MAX > 0xffff
+#define BITS16(a) ((a)&0xffff)
 #define WORD16 unsigned short
+#elif UINT_MAX > 0xffff
+#define BITS16(a) ((a)&0xffff)
+#define WORD16 unsigned int
 #elif ULONG_MAX > 0xffff
 #define WORD16 unsigned long
+#define BITS16(a) ((a)&0xffff)
 #else
 #error No 16 bit value available
 #endif
@@ -79,12 +103,18 @@
 #if USHRT_MAX == 0xffffffff
 #define WORD32 unsigned short
 #define BITS32(a) (a)
+#elif UINT_MAX == 0xffffffff
+#define WORD32 unsigned int
+#define BITS32(a) (a)
 #elif ULONG_MAX == 0xffffffff
 #define WORD32 unsigned long
 #define BITS32(a) (a)
 #elif USHRT_MAX > 0xffffffffUL
 #define WORD32 unsigned short
 #define BITS32(a) ((a)&0xffffffff)
+#elif UINT_MAX > 0xffffffffUL
+#define WORD32 unsigned int
+#define BITS32(a) ((a)&0xffffffff)
 #elif ULONG_MAX > 0xffffffffUL
 #define WORD32 unsigned long
 #define BITS32(a) ((a)&0xffffffff)
@@ -92,12 +122,46 @@
 #error No 32 bit value available
 #endif
 
+/* set WORD64 and BITS64 if we are on a 64 bit machine */
+
+#ifdef ULLONG_MAX
+#if USHRT_MAX == 0xffffffffffffffff
+#define WORD64 unsigned short
+#define BITS64(a) (a)
+#elif UINT_MAX == 0xffffffffffffffff
+#define WORD64 unsigned int
+#define BITS64(a) (a)
+#elif ULONG_MAX == 0xffffffffffffffff
+#define WORD64 unsigned long
+#define BITS64(a) (a)
+#elif ULLONG_MAX == 0xffffffffffffffff
+#define WORD64 unsigned long long
+#define BITS64(a) (a)
+#elif USHRT_MAX > 0xffffffffffffffff
+#define WORD64 unsigned short
+#define BITS64(a) ((a)&0xffffffffffffffff)
+#elif UINT_MAX > 0xffffffffffffffff
+#define WORD64 unsigned int
+#define BITS64(a) ((a)&0xffffffffffffffff)
+#elif ULONG_MAX > 0xffffffffffffffff
+#define WORD64 unsigned long
+#define BITS64(a) ((a)&0xffffffffffffffff)
+#elif ULLONG_MAX > 0xffffffffffffffff
+#define WORD64 unsigned long long
+#define BITS64(a) ((a)&0xffffffffffffffff)
+#endif
+#endif
+
 /* use defined calling conventions for WIN32 */
 
-#ifdef WIN32
-#define FASTCALL  __fastcall
-#ifdef X86ASM
+#ifdef _WIN32
+#define FASTCALL __fastcall
+#ifdef _M_IX86
+#if _M_IX86==300
+#define CRYPTOCALL __fastcall
+#else
 #define CRYPTOCALL __cdecl
+#endif
 #else
 #define CRYPTOCALL __fastcall
 #endif
@@ -105,20 +169,3 @@
 #define FASTCALL
 #define CRYPTOCALL
 #endif
-
-/* assert that WIN32 inline assembler reserved words are not defined */
-
-#ifdef WIN32
-#ifdef BYTE
-#undef BYTE
-#endif
-#ifdef WORD
-#undef WORD
-#endif
-#ifdef DWORD
-#undef DWORD
-#endif
-#ifdef PTR
-#undef PTR
-#endif
-#endif
