@@ -1,8 +1,9 @@
 /*
  * virtual private network daemon (vpnd)
  *
- * cryptographic stuff (c) 1999 Andreas Steinmetz, astmail@yahoo.com
+ * cryptographic stuff (c) 1999 Andreas Steinmetz, ast@domdv.de
  * other code (c) 1999 D.O.M. Datenverarbeitung GmbH, author Andreas Steinmetz
+ * (c) 2001,2005 Andreas Steinmetz
  *
  * License:
  * This code is in the public domain (*) under the GNU public license.
@@ -28,13 +29,14 @@
  * input: anchor - pointer to vpnd global data
  *
  * This procedure performs all necessary processing to transfer data
- * between the local slip interface and the slip interface of the peer.
+ * between the local slip/ppp interface and the slip/ppp interface of
+ * the peer.
  */
 
 void sequencer(VPN *anchor)
 {
 	time_t current;		/* current time in seconds		*/
-	time_t next_slip;	/* next slip retry in seconds		*/
+	time_t next_slip;	/* next slip/ppp retry in seconds	*/
 	time_t next_peer;	/* next peer connect retry in seconds	*/
 	time_t next_ping;	/* next time to idle ping in seconds	*/
 	time_t next_server;	/* next time create srv sk in seconds	*/
@@ -47,7 +49,7 @@ void sequencer(VPN *anchor)
 	WORD32 remotedata;	/* data bytes in peer input buffer	*/
 	WORD08 sendkey;		/* flag that key change is required	*/
 	WORD08 enabled;		/* peer link suspend flag		*/
-	WORD08 slipstate;	/* slip link state and error code	*/
+	WORD08 slipstate;	/* slip/ppp link state and error code	*/
 	WORD08 peerstate;	/* peer link state and error code	*/
 	WORD16 leaky;		/* leaky bucket count			*/
 	WORD16 chklink;		/* flag for link check required		*/
@@ -103,9 +105,9 @@ void sequencer(VPN *anchor)
 				goto perr2;
 			}
 
-			/* create slip interface, handle errors */
+			/* create slip/ppp interface, handle errors */
 
-			if((result=sliphandler(1,anchor))!=0)
+			if((result=anchor->ifhandler(1,anchor))!=0)
 			{
 				logmsg(SLIPFAIL,result,anchor);
 				goto perr2;
@@ -129,14 +131,37 @@ void sequencer(VPN *anchor)
 
 				/* set serial line exception bit */
 
+				FD_ZERO(r);
 				FD_ZERO(e);
 				FD_SET(anchor->serial,e);
 
-				/* wait until select is interrupted or an
-				   exception occurs */
+				/* set current highest descriptor */
 
-				if(select(anchor->serial+1,NULL,NULL,e,NULL)
-					==-1)
+				i=anchor->serial;
+
+				/* if the control socket exists */
+
+				if(ctrl!=-1)
+				{
+					/* check for pending connect request */
+
+					FD_SET(ctrl,r);
+
+					/* always check for exceptions */
+
+					FD_SET(ctrl,e);
+
+					/* set new highest file descriptor
+					   if control is it */
+
+					if(ctrl>i)i=ctrl;
+				}
+
+				/* wait until select is interrupted, an
+				   exception occurs or there's a control
+				   socket connection pending */
+
+				if(select(i+1,r,NULL,e,NULL)==-1)
 				{
 					/* debug message */
 
@@ -152,11 +177,58 @@ void sequencer(VPN *anchor)
 					goto doexit;
 				}
 
+				/* if the control socket exists */
+
+				if(ctrl!=-1)
+				{
+					/* if there's an exception pending */
+
+					if(FD_ISSET(ctrl,e))
+					{
+						/* just close the control
+						   socket, an exception
+						   shouldn't ever happen
+						   here anyway and this
+						   socket isn't required
+						   for processing */
+
+						controlhandler(1);
+					}
+
+					/* if there's a connection request */
+
+					else if(FD_ISSET(ctrl,r))
+					{
+						/* process control request,
+						   handle link down requests */
+
+						if(docontrol(CTRL_ISUP))
+						{
+							/* print log message,
+							   drop interface,
+							   close serial line,
+							   wait and retry */
+
+							logmsg(DCONTROL,0,
+								anchor);
+
+							/* execute linkdown
+							   process, if there
+							   is any */
+
+							doexec(anchor->linkdown,
+								NULL);
+							goto perr3;
+						}
+
+					}
+				}
+
 				/* exception on serial line */
 
 				if(FD_ISSET(anchor->serial,e))
 				{
-					/* print log message, drop slip
+					/* print log message, drop slip/ppp
 					   interface, close serial line,
 					   wait and retry */
 
@@ -172,7 +244,7 @@ void sequencer(VPN *anchor)
 
 			/* error handling */
 
-perr3:			sliphandler(3,anchor);
+perr3:			anchor->ifhandler(3,anchor);
 perr2:			serialhandler(anchor);
 perr1:			sleep(anchor->retry);
 		}
@@ -205,7 +277,7 @@ perr1:			sleep(anchor->retry);
 
 	leaky=anchor->noresponse;
 
-	/* set peer link state and slip link state to down */
+	/* set peer link state and slip/ppp link state to down */
 
 	peerstate=slipstate=0x80;
 
@@ -214,7 +286,7 @@ perr1:			sleep(anchor->retry);
 
 	dosuspend=chklink=next_key=sendkey=localdata=remotedata=0;
 
-	/* force connection to peer, creation of slip link and idle pings */
+	/* force connection to peer, creation of slip/ppp link and idle pings */
 
 	next_ping=next_peer=next_slip=current=next_server=time(NULL)+1;
 
@@ -226,17 +298,30 @@ perr1:			sleep(anchor->retry);
 
 		if(terminate)goto doexit;
 
-		/* if the slip link is down and it's time to retry */
+		/* if the slip/ppp link is down and it's time to retry */
 
 		if(slipstate==0x80&&next_slip<=current)
 		{
-			/* if link creation is successful print info */
+			/* if link creation is successful */
+
+			if(!(slipstate=anchor->ifhandler(0,anchor)))
+			{
+				/* in packetizer mode reset if to packet area */
+
+				if(anchor->pktmode)
+				{
+					anchor->pkt_init(&anchor->ifpkt);
+					pkgcomp_init(&anchor->ifpkt,
+						anchor->cslip);
+				}
+
+				/* print info */
 
-			if(!(slipstate=sliphandler(0,anchor)))
 				logmsg(SLIPUP,0,anchor);
+			}
 		}
 
-		/* handle slip link errors */
+		/* handle slip/ppp link errors */
 
 		if((result=slipstate&0x7f)!=0)
 		{
@@ -244,11 +329,11 @@ perr1:			sleep(anchor->retry);
 
 			logmsg(SLIPFAIL,result,anchor);
 
-			/* drop slip link if it is up */
+			/* drop slip/ppp link if it is up */
 
-			if(anchor->slip)sliphandler(2,anchor);
+			if(anchor->slip)anchor->ifhandler(2,anchor);
 
-			/* set up slip link retry */
+			/* set up slip/ppp link retry */
 
 			next_slip=current+anchor->retry;
 
@@ -256,7 +341,7 @@ perr1:			sleep(anchor->retry);
 
 			remotedata=localdata=0;
 
-			/* set slip state to down */
+			/* set slip/ppp state to down */
 
 			slipstate=0x80;
 		}
@@ -288,7 +373,7 @@ perr1:			sleep(anchor->retry);
 
 			remotedata=localdata=0;
 
-			/* set slip state to down */
+			/* set peer state to down */
 
 			peerstate=0x80;
 		}
@@ -383,6 +468,18 @@ perr1:			sleep(anchor->retry);
 						peerstate=dosuspend=chklink=
 							sendkey=0;
 
+						/* in packetizer mode reset
+						   packet to if area */
+
+						if(anchor->pktmode)
+						{
+							anchor->pkt_init(
+								&anchor->pktif);
+							pkgcomp_init(
+								&anchor->ifpkt,
+								anchor->cslip);
+						}
+
 						/* execute linkup process,
 						   if there is any */
 
@@ -440,7 +537,7 @@ perr1:			sleep(anchor->retry);
 		FD_ZERO(w);
 		FD_ZERO(e);
 
-		/* if the slip link is up */
+		/* if the slip/ppp link is up */
 
 		if(!slipstate)
 		{
@@ -450,7 +547,12 @@ perr1:			sleep(anchor->retry);
 
 			/* check for write possible if peer data available */
 
-			if(remotedata)FD_SET(anchor->pty,w);
+			if(anchor->pktmode)
+			{
+				if(anchor->pktif.state==STREAM)
+					FD_SET(anchor->pty,w);
+			}
+			else if(remotedata)FD_SET(anchor->pty,w);
 
 			/* always check for exceptions */
 
@@ -476,7 +578,12 @@ perr1:			sleep(anchor->retry);
 
 			/* check for write possible if local data available */
 
-			if(localdata||sendkey)FD_SET(anchor->peer,w);
+			if(anchor->pktmode)
+			{
+				if(anchor->ifpkt.state==STREAM||sendkey)
+					FD_SET(anchor->peer,w);
+			}
+			else if(localdata||sendkey)FD_SET(anchor->peer,w);
 
 			/* always check for exceptions */
 
@@ -488,7 +595,7 @@ perr1:			sleep(anchor->retry);
 		}
 
 		/* if the peer link is down in server mode and if the
-		   sever socket exists */
+		   server socket exists */
 
 		else if(!anchor->mode&&anchor->server!=-1)
 		{
@@ -505,6 +612,23 @@ perr1:			sleep(anchor->retry);
 			if(anchor->server>i)i=anchor->server;
 		}
 
+		/* if the control socket exists */
+
+		if(ctrl!=-1)
+		{
+			/* check for pending connect request */
+
+			FD_SET(ctrl,r);
+
+			/* always check for exceptions */
+
+			FD_SET(ctrl,e);
+
+			/* set new highest file descriptor if control is it */
+
+			if(ctrl>i)i=ctrl;
+		}
+
 		/* set 1 second timeout for select call */
 
 		t.tv_sec=1;
@@ -608,6 +732,18 @@ perr1:			sleep(anchor->retry);
 						peerstate=dosuspend=chklink=
 							sendkey=0;
 
+						/* in packetizer mode reset
+						   packet to if area */
+
+						if(anchor->pktmode)
+						{
+							anchor->pkt_init(
+								&anchor->pktif);
+							pkgcomp_init(
+								&anchor->ifpkt,
+								anchor->cslip);
+						}
+
 						/* execute linkup process
 						   if there is any */
 
@@ -665,6 +801,7 @@ perr1:			sleep(anchor->retry);
 					disconnect(0,anchor);
 					peerstate=0x80;
 					enabled=0;
+					remotedata=localdata=0;
 				}
 			}
 
@@ -709,6 +846,73 @@ perr1:			sleep(anchor->retry);
 
 		next_ping=current+anchor->keepalive;
 
+		/* if the control socket exists */
+
+		if(ctrl!=-1)
+		{
+			/* if there's an exception pending */
+
+			if(FD_ISSET(ctrl,e))
+			{
+				/* just close the control socket, an exception
+				   shouldn't ever happen here anyway and this
+				   socket isn't required for processing */
+
+				controlhandler(1);
+			}
+
+			/* if there's a connection request */
+
+			else if(FD_ISSET(ctrl,r))
+			{
+				/* process the control request */
+
+				i=docontrol(peerstate?CTRL_ISDOWN:CTRL_ISUP);
+
+				/* if the peer link exists and a link down
+				   request has to be processed */
+
+				if(!peerstate&&i)
+				{
+					/* disconnect from peer in requested
+					   mode */
+
+					disconnect(i==CTRL_FORCEDOWN?4:0,
+						anchor);
+
+					/* flush data buffers */
+
+					remotedata=localdata=0;
+
+					/* mark peer link as down */
+
+					peerstate=0x80;
+
+					/* print proper message according
+					   to mode, in client mode retry
+					   and take care of suspend */
+
+					if(!anchor->mode&&!anchor->serialmode)
+					{
+						logmsg(DSCONTROL,0,anchor);
+						next_server=current+1;
+					}
+					else
+					{
+						logmsg(DCONTROL,0,anchor);
+						next_peer=current+anchor->retry;
+						if(anchor->mode&&
+							anchor->suspend)
+								enabled=0;
+					}
+
+					/* restart main loop */
+
+					continue;
+				}
+			}
+		}
+
 		/* if the peer connection exists and there's a peer exception */
 
 		if(!peerstate&&FD_ISSET(anchor->peer,e))
@@ -799,6 +1003,18 @@ perr1:			sleep(anchor->retry);
 
 						disconnect(1,anchor);
 
+						/* in packetizer mode reset
+						   packet to if area */
+
+						if(anchor->pktmode)
+						{
+							anchor->pkt_init(
+								&anchor->pktif);
+							pkgcomp_init(
+								&anchor->ifpkt,
+								anchor->cslip);
+						}
+
 						/* execute linkup process,
 						   if there is any */
 
@@ -813,8 +1029,8 @@ perr1:			sleep(anchor->retry);
 			}
 		}
 
-		/* if the slip link exists and there's an exception pending
-		   set slip error code */
+		/* if the slip/ppp link exists and there's an exception pending
+		   set slip/ppp error code */
 
 		if(!slipstate&&FD_ISSET(anchor->pty,e))slipstate=229;
 
@@ -915,7 +1131,7 @@ perr1:			sleep(anchor->retry);
 			if((i=datarecv(rptr,MAXBLOCK,anchor))>0)
 			{
 				/* store new received data length 
-				   if slip link is up */
+				   if slip/ppp link is up */
 
 				if(!slipstate)remotedata=i-1;
 
@@ -933,13 +1149,81 @@ perr1:			sleep(anchor->retry);
 			else peerstate=233;
 		}
 
-		/* if there's data from the peer in the buffer and if we can
-		   write to the slip interface */
+		/* in packetizer mode */
 
-		if(!slipstate&&remotedata&&FD_ISSET(anchor->pty,w))
+		if(anchor->pktmode)
+		{
+			/* if there is a frame in the packet buffer and if we
+			   can write to the slip/ppp interface */
+
+			if(anchor->pktif.state==STREAM&&!slipstate&&
+				FD_ISSET(anchor->pty,w))
+			{
+				/* if write to slip/ppp is successful */
+
+				if((i=write(anchor->pty,anchor->pktif.data,
+					anchor->pktif.length))>0)
+				{
+					/* adjust data pointer */
+
+					anchor->pktif.data+=i;
+
+					/* adjust data length */
+
+					anchor->pktif.length-=i;
+
+					/* if the frame was completely sent
+					   reset the paketizer work area */
+
+					if(!anchor->pktif.length)
+						anchor->pkt_init(
+							&anchor->pktif);
+				}
+
+				/* in case of write failure  set slip/ppp
+				   error code and reset packetizer work area */
+
+				else
+				{
+					slipstate=230;
+					anchor->pkt_init(&anchor->pktif);
+				}
+			}
+
+			/* if there is no frame in the packet buffer and we
+			   have data available */
+
+			if(anchor->pktif.state!=STREAM&&remotedata)
+			{
+				/* depacketize as much data as possible */
+
+				i=anchor->pkt_to_if(&anchor->pktif,rptr,
+					remotedata,anchor->recode);
+
+				/* adjust data pointer */
+
+				rptr+=i;
+
+				/* adjust data length */
+
+				remotedata-=i;
+			}
+
+			/* if there is a frame in the packet buffer and the
+			   slip/ppp interface is down reset packetizer work
+			   area */
+
+			if(anchor->pktif.state==STREAM&&slipstate)
+				anchor->pkt_init(&anchor->pktif);
+		}
+
+		/* if there's data from the peer in the buffer and if we can
+		   write to the slip/ppp interface */
+
+		else if(!slipstate&&remotedata&&FD_ISSET(anchor->pty,w))
 		{
 
-			/* if write to slip is successful */
+			/* if write to slip/ppp is successful */
 
 			if((i=write(anchor->pty,rptr,remotedata))>0)
 			{
@@ -952,13 +1236,13 @@ perr1:			sleep(anchor->retry);
 				remotedata-=i;
 			}
 
-			/* in case of write failure  set slip error code */
+			/* in case of write failure  set slip/ppp error code */
 
 			else slipstate=230;
 		}
 
-		/* if data from slip are available and if there's no data
-		   from slip in the buffer */
+		/* if data from slip/ppp are available and if there's no data
+		   from slip/ppp in the buffer */
 
 		if(!slipstate&&!localdata&&FD_ISSET(anchor->pty,r))
 		{
@@ -979,20 +1263,72 @@ perr1:			sleep(anchor->retry);
 
 			lptr=anchor->lbfr;
 
-			/* in case of read failure set slip error code */
+			/* in case of read failure set slip/ppp error code */
 
 			if(!(i=read(anchor->pty,lptr,MAXBLOCK)))slipstate=231;
 
 			/* in case of successful read adjust buffer length
-			   if peer link exists */
+			   if peer link exists or in packetizer mode */
 
-			else if(!peerstate)localdata=i;
+			else if(!peerstate||anchor->pktmode)localdata=i;
 		}
 
-		/* if there is data from the slip interface available and
+		/* in packetizer mode */
+
+		if(anchor->pktmode)
+		{
+			/* if there is a frame in the packet buffer and if we
+			   can write to the peer */
+
+			if(anchor->ifpkt.state==STREAM&&!peerstate&&
+				FD_ISSET(anchor->peer,w))
+			{
+				/* in case of successful send reset
+				   suspend and link test counters */
+
+				if(datasend(anchor->ifpkt.data,
+					anchor->ifpkt.length,anchor))
+					dosuspend=chklink=0;
+
+				/* in case of send failure set peer error */
+
+				else peerstate=234;
+
+				/* reset packetizer work area */
+
+				anchor->pkt_init(&anchor->ifpkt);
+			}
+
+			/* while there is no frame in the packet buffer and we
+			   have data available */
+
+			while(anchor->ifpkt.state!=STREAM&&localdata)
+			{
+				/* depacketize as much data as possible */
+
+				i=anchor->if_to_pkt(&anchor->ifpkt,lptr,
+					localdata,anchor->recode);
+
+				/* adjust data pointer */
+
+				lptr+=i;
+
+				/* adjust data length */
+
+				localdata-=i;
+			}
+
+			/* if there is a frame in the packet buffer and the
+			   peer link is down reset packetizer work area */
+
+			if(anchor->ifpkt.state==STREAM&&peerstate)
+				anchor->pkt_init(&anchor->ifpkt);
+		}
+
+		/* if there is data from the slip/ppp interface available and
 		   if data can be written to peer */
 
-		if(!peerstate&&localdata&&FD_ISSET(anchor->peer,w))
+		else if(!peerstate&&localdata&&FD_ISSET(anchor->peer,w))
 		{
 			/* flush buffer in case of successful send, reset
 			   suspend and link test counters */
